import argparse
import logging
import sys
from .outputs import LoggingConfig, PackageLogFilter
from .oauth import OAuth2, GoogleOAuth
from .outputs import Print
logger = logging.getLogger(__name__)


class Program:
    """Contains static methods for sub-commands to start the processing program with args

    The name of the methods are the same as the name of the sub parsers in subparsers.add_parser()
    """
    @staticmethod
    def oauth(args):
        auth = OAuth2(
            client_id=args.id, 
            client_secret=args.secret, 
            auth_endpoint=args.auth_url, 
            token_endpoint=args.token_url
        )
        if args.redirect:
            auth_url = auth.authentication_url(scope=args.scope, redirect_uri=args.redirect)
        else:
            auth_url = auth.authentication_url(scope=args.scope)
        Print.blue("Please visit the following URL to authenticate:")
        print(auth_url)
        print()
        code = input("Enter the code you get after authentication:")
        if args.redirect:
            response = auth.exchange_token(code, redirect_uri=args.redirect)
        else:
            response = auth.exchange_token(code)
        Print.green("OAuth Tokens:")
        Print.print(response)

    @staticmethod
    def auth_google(args):
        """Authenticate with Google APIs
        """
        auth = GoogleOAuth(client_id=args.id, client_secret=args.secret)
        auth_url = auth.authentication_url(scope=args.scope)
        Print.blue("Please visit the following URL to authenticate with Google:")
        print(auth_url)
        print()
        code = input("Enter the code you get after authentication:")
        response = auth.exchange_token(code)
        Print.green("OAuth Tokens:")
        Print.print(response)


def main():
    parser = argparse.ArgumentParser(description="Entry points to Aries command line tools.")
    subparsers = parser.add_subparsers(title="Program", help="Program", dest='program')

    sub_parser = subparsers.add_parser(
        "oauth",
        help="Authenticate with OAuth 2.0 to obtain access token and refresh token."
    )
    sub_parser.add_argument('--auth_url', type=str, required=True, help="OAuth authentication endpoint")
    sub_parser.add_argument('--token_url', type=str, required=True, help="OAuth token endpoint")
    sub_parser.add_argument('--id', type=str, required=True, help="OAuth client ID")
    sub_parser.add_argument('--secret', type=str, required=True, help="OAuth client secret")
    sub_parser.add_argument('--scope', type=str, required=True, nargs='+', help="OAuth Scopes")
    sub_parser.add_argument('--redirect', type=str, help="Redirect URL")

    sub_parser = subparsers.add_parser(
        "auth_google",
        help="Authenticate with Google to obtain access token and refresh token."
    )
    sub_parser.add_argument('--id', type=str, required=True, help="Google OAuth client ID")
    sub_parser.add_argument('--secret', type=str, required=True, help="Google OAuth client secret")
    sub_parser.add_argument('--scope', type=str, required=True, nargs='+', help="OAuth 2.0 Scopes for Google APIs")

    # Parse command
    args = parser.parse_args()
    # Show help if no sub parser matched.
    if not vars(args).keys() or not args.program or not hasattr(Program, args.program):
        parser.parse_args(["-h"])
        return

    # Run program
    func = getattr(Program, args.program)
    func(args)
    sys.exit(0)


if __name__ == '__main__':
    with LoggingConfig(filters=[PackageLogFilter(packages="Aries")]):
        main()



import sys


def append_sys_path(folder_path):
    """Appends a path to sys.path
    """
    if folder_path not in sys.path:
        sys.path.append(folder_path)


def group_by_attr(records, attr):
    """Groups a list of objects by the value of an attribute

    Returns: A dictionary, where
        The keys are the distinct values of the given attribute.
        Each value is a list of objects with the same attribute value.
    """
    groups = dict()
    for record in records:
        key = getattr(record, attr)
        group = groups.get(key, [])
        group.append(record)
        groups[key] = group
    return groups



import requests
import logging
logger = logging.getLogger(__name__)


class OAuth2:
    def __init__(self, client_id, client_secret, auth_endpoint, token_endpoint) -> None:
        super().__init__()
        self.auth_endpoint = auth_endpoint
        self.token_endpoint = token_endpoint
        self.client_id = client_id
        self.client_secret = client_secret
        self.scope = None

    def authentication_url(self, scope, redirect_uri="urn:ietf:wg:oauth:2.0:oob", response_type="code", **kwargs):
        """Generates the authentication URL for user to authenticate

        Args:
            scope (str or list): Scopes/Permissions to be requested.
            redirect_uri: The redirect URI after authentication.
                The default redirect_uri is a special value,
                which means to display the authentication code in the browser.

        Returns: A URL for user to authenticate.

        """
        self.scope = scope
        if isinstance(scope, list):
            scope = "%20".join(scope)

        auth_url = "%s?scope=%s&" \
                   "redirect_uri=%s&" \
                   "client_id=%s&" \
                   "response_type=%s" % \
                   (
                       self.auth_endpoint,
                       scope,
                       redirect_uri,
                       self.client_id,
                       response_type
                   )
        for key, val in kwargs.items():
            auth_url += "&%s=%s" % (key, val)
        return auth_url

    def exchange_token(self, auth_code, redirect_uri="urn:ietf:wg:oauth:2.0:oob", scope=None):
        data = {
            "code": auth_code,
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "redirect_uri": redirect_uri,
            "grant_type": "authorization_code"
        }

        if scope:
            data["scope"] = scope
        elif self.scope:
            data["scope"] = self.scope

        response = requests.post(self.token_endpoint, data).json()
        return response

    def refresh_access_token(self, refresh_token, scope=None):
        data = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "refresh_token": refresh_token,
            "grant_type": "refresh_token"
        }

        if scope:
            data["scope"] = scope
        elif self.scope:
            data["scope"] = self.scope

        response = requests.post(self.token_endpoint, data).json()
        if response.get("error"):
            logger.error("Error when getting access token: %s" % response.get("error"))
        return response.get("access_token")


class GoogleOAuth:
    """
    See Also:
        Using OAuth 2.0 to Access Google APIs
        https://developers.google.com/identity/protocols/oauth2

    """

    token_endpoint = "https://www.googleapis.com/oauth2/v4/token"
    auth_endpoint = "https://accounts.google.com/o/oauth2/v2/auth"

    def __init__(self, client_id, client_secret):
        """
        OAuth 2.0 client IDs can be obtained at Google API Console:
        https://console.developers.google.com/

        Args:
            client_id: Google OAuth client ID
            client_secret: Google OAuth client secret
        """
        self.client_id = client_id
        self.client_secret = client_secret

    def authentication_url(self, scope, redirect_uri="urn:ietf:wg:oauth:2.0:oob"):
        """Generates the authentication URL for user to authenticate with Google

        Args:
            scope (str or list): OAuth 2.0 Scopes for Google APIs.
            redirect_uri: The redirect URI for Google to redirect after authentication.
                The default redirect_uri is a special value,
                which will cause Google to display the authentication code in the browser.

        Returns: A URL for user to authenticate with Google.

        See Also:
            OAuth 2.0 Scopes for Google APIs
            https://developers.google.com/identity/protocols/oauth2/scopes

        """
        if isinstance(scope, list):
            scope = "&".join(["scope=%s" % s for s in scope])
        else:
            scope = "scope=%s" % scope

        auth_url = "%s?%s&" \
                   "access_type=offline&" \
                   "include_granded_scopes=true&" \
                   "redirect_uri=%s&" \
                   "client_id=%s&" \
                   "response_type=code" % \
                   (
                       self.auth_endpoint,
                       scope,
                       redirect_uri,
                       self.client_id
                   )
        return auth_url

    def exchange_token(self, auth_code, redirect_uri="urn:ietf:wg:oauth:2.0:oob"):
        response = requests.post(GoogleOAuth.token_endpoint, {
            "code": auth_code,
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "redirect_uri": redirect_uri,
            "grant_type": "authorization_code"
        }).json()
        return response

    def refresh_access_token(self, refresh_token):
        response = requests.post(GoogleOAuth.token_endpoint, {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "refresh_token": refresh_token,
            "grant_type": "refresh_token"
        }).json()
        if response.get("error"):
            logger.error("Error when getting access token: %s" % response.get("error"))
        return response.get("access_token")



"""Build Script for setuptools

This build script must be executed outside of the Aries directory.

See Also: https://packaging.python.org/tutorials/packaging-projects/
"""
import setuptools
from Aries.setup import setup


setuptools.setup(
    name="Aries-core",
    author="Qiu Qin",
    author_email="qiuosier@gmail.com",
    description="Python package providing shortcuts to small tasks like string manipulation, "
                "running background tasks, configuring logging, accessing web API etc.",
    url="https://github.com/qiuosier/Aries",
    version=setup.get_version(),
    long_description=setup.get_description("Aries/README.md"),
    install_requires=setup.get_requirements("Aries/setup/requirements_core.txt"),
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(exclude=(
        "Aries.setup",
        "Aries.gcp",
        "Aries.statistics",
        "Aries.storage",
        "Aries.tests",
        "Aries.visual",
    )),
    package_data={
        "Aries": ["assets/*"]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



"""Contains classes for reading table-like data.
"""
import csv
import logging
logger = logging.getLogger(__name__)


def list_index(values):
    return {values[i]: i for i in range(len(values))}


class TableRow(list):
    """Represents a row in a table as a list
    """
    def __init__(self, values, headers=None, index=None):
        """Initializes a list of values as a row in a table.

        Args:
            values (list): A list of values, usually strings.
            headers (list): Optional. A list as the header of the table.
            index (dict): Optional. A dictionary containing the mapping from the headers to the indices.
                key: a column header in headers
                value: 0-based index of header in headers.

            headers and index are used for accessing the value using the get(header) method.
            headers will be ignored if index is given.
            index will be generated from headers if it is not given.
        """
        if index is not None:
            self.headers = list(index.keys())
            self.index = index
        elif headers is not None:
            self.headers = headers
            self.index = list_index(headers)
        else:
            self.headers = []
            self.index = dict()
        self.upper_index = {str(k).upper(): v for k, v in self.index.items()}
        super().__init__(values)

    def get(self, header, default=None, case_sensitive=True):
        if case_sensitive:
            col_index = self.index.get(header)
        else:
            col_index = self.upper_index.get(str(header).upper())
        if col_index is not None and col_index < len(self):
            return self[col_index]
        else:
            return default

    def keys(self):
        return self.headers


class TableData:
    """Represents a table-like data set, e.g. data from a CSV file.
    """
    def __init__(self, headers, data=None):
        """Initializes a table with headers and data.

        Args:
            headers: A list of strings as the headers of a table
            data: A 2-D list of the data in the table.
                data[i] is a list containing the data of the ith row.
        """
        self.headers = headers
        self.data = data
        self.header_index = list_index(headers)

    def get_header_index(self, header):
        """Gets the index of a header in the headers of the table.

        Args:
            header: one of the header in self.headers

        Returns: The 0-based index of the header, -1 if the header is not found.

        """
        return self.header_index.get(header, -1)

    def get_value(self, row_index, header, default=None):
        """Gets the value of a particular cell in the table.

        Args:
            row_index: 0-based row index
            header: The header of the column as it appears in self.headers
            default: The default value to be returned if the column is missing.

        Returns:

        """
        col_index = self.get_header_index(header)
        row = self.data[row_index]
        if col_index < len(row):
            return row[col_index]
        else:
            return default

    def __len__(self):
        """The number of rows in the table
        """
        return len(self.data)

    def __iter__(self):
        """Resets the row pointer to point to the first row.
        """
        self.row_pointer = 0
        return self

    def __next__(self):
        """Gets the next row of data from the table.

        Returns: A TableRow list.

        """
        if self.row_pointer < len(self.data):
            row = TableRow(self.data[self.row_pointer], self.headers)
            self.row_pointer += 1
            return row
        else:
            raise StopIteration

    @staticmethod
    def infer_csv_dialect(lines):
        logger.debug("Detecting dialect in CSV...")
        counter = 0

        # buffer = ""
        # for line in lines:
        #     buffer += line
        #     counter += 1
        #     if counter > 20:
        #         break
        # dialect = csv.Sniffer().sniff(buffer, '\t')

        for dialect in [csv.excel_tab, csv.excel]:
            delimiter_counts = dict()
            for line in lines:
                c = str(line).count(dialect.delimiter)
                if c == 0:
                    continue
                delimiter_counts[c] = delimiter_counts.get(c, 0) + 1
                counter += 1
                if counter > 20:
                    break
            if not delimiter_counts:
                continue
            if max(delimiter_counts.values()) > counter * 0.5:
                break
        logger.debug(repr("Detected delimiter: %s" % dialect.delimiter))
        return dialect

    @staticmethod
    def __parse_csv(lines, header_row=0, **kwargs):
        """Initializes a table from text lines in CSV format.

        Args:
            lines: A iterable object returning rows as strings in CSV format.
            header_row: The 0-based index of the header row in the file. Defaults to 0.
                rows before the header row will be ignored.
                values in the header row will be stored in self.header as a list.
                rows after the header row will be stored in self.data
            **kwargs: keyword arguments for python csv.reader()

        Returns: A TableData object.

        """
        headers = []
        data = []
        # Determine the delimiter if it is not given
        if "delimiter" not in kwargs and "dialect" not in kwargs:
            dialect = TableData.infer_csv_dialect(lines)
            kwargs["dialect"] = dialect
        csv_data = csv.reader(lines, **kwargs)
        for i, row in enumerate(csv_data):
            # Skip the rows before the header row
            if i < header_row:
                continue
            # Header row
            if i == header_row:
                headers = row
            else:
                # Data row
                data.append(row)
        return TableData(headers, data)

    @staticmethod
    def from_csv(file_path, header_row=0, **kwargs):
        """Initializes the table from a CSV file.

        Args:
            file_path: The full file path.
            header_row: The 0-based index of the header row in the file. Defaults to 0.
                rows before the header row will be ignored.
                values in the header row will be stored in self.header as a list.
                rows after the header row will be stored in self.data
            **kwargs: keyword arguments for python csv.reader()

        Returns: A TableData object.

        """
        # TODO: open file with URI
        with open(file_path, 'r') as f:
            return TableData.__parse_csv(f, header_row, **kwargs)

    @staticmethod
    def from_csv_string(s, header_row=0, **kwargs):
        """

        Args:
            s (str): A text string containing all rows of a CSV file.
            header_row: The 0-based index of the header row in the file. Defaults to 0.
                rows before the header row will be ignored.
                values in the header row will be stored in self.header as a list.
                rows after the header row will be stored in self.data
            **kwargs: keyword arguments for python csv.reader()

        Returns:

        """
        return TableData.__parse_csv(s.splitlines(), header_row, **kwargs)


class TableCSVFile(TableData):
    def __init__(self, file_path, header_row=0, **kwargs):
        self.file_path = file_path
        self.header_row = header_row
        self.kwargs = kwargs
        # headers and header_index will be initialized by TableData.__init___()
        self.headers = None
        self.header_index = None
        headers = self.__sniff()
        TableData.__init__(self, headers)

    def __sniff(self):
        """
        """
        with open(self.file_path, 'r') as f:
            # Determine the delimiter if it is not given
            if "delimiter" not in self.kwargs and "dialect" not in self.kwargs:
                dialect = TableData.infer_csv_dialect(f)
                self.kwargs["dialect"] = dialect
        # Find Header row
        for i, row in enumerate(self.rows):
            if i == self.header_row:
                return row

    def __len__(self):
        """The number of data rows in the table
        """
        return len(list(self.rows))

    def __iter__(self):
        """
        """
        return self.rows

    @property
    def rows(self):
        with open(self.file_path, 'r') as f:
            csv_data = csv.reader(f, **self.kwargs)
            for i, row in enumerate(csv_data):
                # Skip the rows before the header row
                if i < self.header_row:
                    continue
                yield TableRow(row, self.headers, index=self.header_index)
